package com.uas.search.console.b2b.jms;

import com.alibaba.fastjson.JSONException;
import com.alibaba.fastjson.JSONObject;
import com.uas.search.b2b.exception.SearchException;
import com.uas.search.b2b.service.SearchService.Table_name;
import com.uas.search.console.b2b.util.ClassAndTableNameUtils;
import com.uas.search.util.StringUtils;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.stereotype.Service;

/**
 * 对得到的队列消息进行解析的工具
 *
 * @author sunyj
 * @since 2016年11月9日 上午11:30:02
 */
@Service
public class QueueMessageParser {

    /**
     * 对得到的json消息进行解析
     *
     * @param tableName  表名
     * @param dataId     数据 id
     * @param methodType 更改类型
     * @param data       数据
     * @return ParsedQueueMessage对象
     * @throws JSONException
     */
    public ParsedQueueMessage parse(String tableName, Long dataId, String methodType, String data) throws JSONException {
		if (StringUtils.isEmpty(tableName) || StringUtils.isEmpty(methodType)) {
            throw new IllegalArgumentException("tableName 或 methodType 不合法");
		}
		// id 或 data 必须有一个存在
		if (dataId == null && StringUtils.isEmpty(data)) {
            throw new IllegalArgumentException("dataId 和 data 不合法");
		}

		ParsedQueueMessage parsedQueueMessage = new ParsedQueueMessage();
		// 解析数据库表的更改类型
		if (methodType.equalsIgnoreCase("insert")) {
			parsedQueueMessage.setMethodType(ParsedQueueMessage.INSERT);
		}

		else if (methodType.equalsIgnoreCase("update")) {
			parsedQueueMessage.setMethodType(ParsedQueueMessage.UPDATE);
		}

		else if (methodType.equalsIgnoreCase("delete")) {
			parsedQueueMessage.setMethodType(ParsedQueueMessage.DELETE);
		} else {
            throw new IllegalStateException("unsupported method type: " + methodType);
		}

		// 解析哪个表有更改
		Class<?> clazz = ClassAndTableNameUtils.toClass(Table_name.valueOf(tableName.toUpperCase()));
        Object object;
		if (dataId != null) {
            object = parseById(dataId, clazz, methodType);
		} else {
            object = parseByData(data, clazz);
		}

		parsedQueueMessage.setObject(object);
		return parsedQueueMessage;
	}

    /**
     * 根据消息队列的消息解析对象
     *
     * @param <T>
     * @param data       数据
     * @param clazz      实体类型
     * @return 实体对象
     */
    private <T> T parseByData(String data, Class<T> clazz) {
        return JSONObject.parseObject(data, clazz);
	}

	/**
	 * 根据消息队列的消息解析对象
	 *
	 * @param <T>
	 *
     * @param dataId     数据 id
     * @param clazz      实体类型
     * @param methodType 更改类型
     * @return 实体对象
	 */
	private <T> T parseById(Long dataId, Class<T> clazz, String methodType) {
		// 获取实体的dao
		JpaRepository<T, Long> dao = ClassAndTableNameUtils.getDao(clazz);
		// delete操作
		// 删除后数据库中可能已经没有相应数据了，无法通过dao获取，需要手动创建对象
		if (methodType.equalsIgnoreCase("delete")) {
			return ClassAndTableNameUtils.createObject(clazz, dataId);
		} else {
			try {
				return dao.findOne(dataId);
			} catch (Exception e) {
				// 防止SQLRecoverableException导致应用终止
				throw new SearchException(e).setDetailedMessage(e);
			}
		}
	}
}
